html, body {
    width: 100%;
    height: 100%;
    margin: 0px;
    border: 0;
    overflow: hidden; /*  Disable scrollbars */
    display: block;  /* No floating content on sides */
}

.jsPanel-hdr-r {
    display: none;
}

.ui-button-icon-only {
    padding-top: 6px;
    padding-bottom: 6px;
}

.jsPanel-content {
    padding: 4px;
    overflow-y: visible !important;
}

.jsPanel-hdr img {
    padding-top: 4px;
    padding-bottom: 4px;
    display: block;
    margin-left: auto;
    margin-right: auto;
}

.ui-slider .ui-slider-handle {
    height: 20px;
}

.ui-icon-custom-default { 
    background-image: url(images/default.png) !important; 
}

.ui-icon-custom-orthanc { 
    background-image: url(images/orthanc-icon.png) !important; 
}

.ui-icon-custom-stretch { 
    background-image: url(images/stretch.png) !important; 
}

.ui-icon-custom-bone { 
    background-image: url(images/bone.png) !important; 
}

.ui-icon-custom-lung { 
    background-image: url(images/lung.png) !important; 
}

.ui-icon-custom-interpolation { 
    background-image: url(images/interpolation.png) !important; 
}

.ui-icon-custom-inversion { 
    background-image: url(images/inversion.png) !important; 
}

#dicomImageWrapper {
    top: 0px;
    left: 0px;
    width: 100%;
    height: 100%;
    position: absolute;
    background-color:  black;
    color: white;
    overflow: hidden; 
    font-family:  Arial, Helvetica, sans-serif;
}

#dicomImage {
    width: 100%;
    height: 100%;
    top: 0px;
    left: 0px; 
    position: absolute;
}

#topleft {
    position: absolute;
    top: 10px; 
    left: 10px; 
}

#topleft img {
    vertical-align: middle;
}

#topright {
    position: absolute;
    top: 10px; 
    right: 10px; 
    text-align: right;
}

#bottomright {
    position: absolute;
    bottom:25px; 
    right:10px;
}

#bottomleft {
    position: absolute;
    bottom: 25px; 
    left: 10px;
}

#bottomcenter {
    position: absolute;
    bottom: 25px; 
    left: 10px;
    right: 10px; 
    width: 100%;
    text-align: center;
}

#bottom {
    position: absolute;
    bottom: 0px;
    left: 0px;
    height: 20px;
    width: 100%;
}

.alert {
    color: #f00;
    font-weight: bold;
}

